import contextlib
import pyodbc
import re
import logging
from . import exceptions
from ._regex_lookup_dict import RegexLookupDict

logger = logging.getLogger(__name__)
CONNECTION_FAILED_ERRORS = RegexLookupDict(
    {
        "login failed for user": exceptions.ConnectionFailed(
            "Failed to log in to the default database provided. Please check the user has access to "
            "the default database."
        ),
        "the login is from an untrusted domain and cannot be used with integrated": exceptions.ConnectionFailed(
            "Failed to log in to the Microsoft SQL Server. Please check the credentials provided."
        ),
        "adaptive server is unavailable or does not exist": exceptions.ConnectionFailed(
            "Unable to connect to the address and port provided. Please check the address or port provided."
        ),
    }
)

EXECUTE_ERRORS = RegexLookupDict(
    {
        "invalid object name": exceptions.InvalidSQLQuery(
            "An invalid object name was used in the provided SQL Query. "
            "Please check the table names and schema supplied."
        ),
        "invalid column name": exceptions.InvalidSQLQuery("An invalid column name was used in the provided SQL Query."),
        "the select permission was denied": exceptions.PermissionDenied(
            "The user provided for the server connection does not have read access"
            "to all the tables specified in the SQL Query."
        ),
        "not able to access the database": exceptions.PermissionDenied(
            "The user provided for the server connection does not have read access to the database specified."
        ),
        r"the sql contains \d+ parameter markers, but \d+ parameters were supplied": exceptions.InvalidSQLQuery(
            "Not enough values were supplied for all placeholders in the SQL Query."
        ),
    }
)


@contextlib.contextmanager
def _connect(
    *,
    address: str,
    instance: str,
    default_database: str,
    username: str,
    password: str,
    port=1433,
    freetds_driver="/usr/lib/libtdsodbc.so.0",
):
    connection = None

    try:
        connection = pyodbc.connect(
            f"DRIVER={{{freetds_driver}}};"
            f"SERVER={address}\\{instance};PORT={port};DATABASE={default_database};UID={username};PWD={password};"
            "TDS_VERSION=8.0"
        )
        yield connection
    except (pyodbc.ProgrammingError, pyodbc.Error) as e:
        error_message = str(e)
        logger.exception(error_message)
        try:
            raise CONNECTION_FAILED_ERRORS[error_message.lower()]
        except KeyError:
            raise exceptions.ConnectionFailed(
                "Failed to connect to the Microsoft SQL Server. Please check the logs for more info."
            )
    finally:
        if connection:
            connection.close()


def _execute(connection, *, query: str, values: tuple = ()):
    try:
        cursor = connection.cursor()
        cursor.execute(query, values)
    except (pyodbc.ProgrammingError, pyodbc.Error) as e:

        # Exceptions from pyodbc contain an error code and message. Unfortunately, some error codes are reused so we
        # need to check the message
        error_message = str(e)
        logger.exception(error_message)
        try:
            raise EXECUTE_ERRORS[error_message.lower()]
        except KeyError:
            if "The driver did not supply an error!" in error_message and "?" in query and not values:
                raise exceptions.InvalidSQLQuery("No values were provided for the placeholders in the SQL Query.")
            else:
                raise exceptions.InvalidSQLQuery("Executing SQL Query failed. Please check the logs.")
    return cursor


def _check_for_sql_injection(query, exception_message):
    delimiter_count = query.count(";")
    if delimiter_count == 1 and not query.endswith(";"):
        raise exceptions.SQLInjection(exception_message)
    elif delimiter_count > 1:
        raise exceptions.SQLInjection(exception_message)


def _remove_comments(query):
    # Remove comments from SQL query for our SQLInjection checks
    query = re.sub(r"/\*[^*]*(?:\*(?!/)[^*]*)*\*/", "", query, flags=re.MULTILINE)  # Remove multi line comments
    # Remove -- comments
    lines = query.splitlines()
    return " ".join([line.split("--")[0].strip() for line in lines])
