from typeguard import typechecked
from typing import List
from . import _read
from ._records import _generate_records, Record
from ._read import SantisedReadQuery


@typechecked
def sanitise_query(sql_query: str) -> SantisedReadQuery:
    """
    Sanitises a SQL query to be provided to the :py:func:`read.query`

    :param query: The SQL query to be sanitised.

    :return: A SantisedReadQuery

    :raises InvalidSQL: if the query is not a read operation
    :raises SQLInjection: if a possible SQL Injection attempt has been found

    """
    return _read._sanitise(sql_query)


@_generate_records()
def query(connection, *, sql_query: str, values: tuple = ()) -> List[Record]:  # type: ignore
    """
    Executes a read operation query against the MSSQL Server.

    :param connection: An open MSSQL Connection obtained via :py:func:`client.initialize`.
    :param sql_query: An SQL query that has been sanitised via :py:func:`read.sanitise_query`.
    :param values: Optional. A tuple of values to replace any ? placeholders in the query.
        This helps to prevent any SQL injection attempts if input is provided from untrusted sources.

    :return: A list of Records

    :raises UnhandledRecordData: if the query returned unexpected data

    .. warning::
        Any user input should be handled with extreme care as it may contain SQL injection attempts to
        read, update or delete data from the database. We recommend using the '?' placeholders for any user
        input in the query and passing the user input as a tuple to the `values` parameter.

        Example:

        .. code-block:: python

            user_name = ppa.input_text("Please enter your name")

            query = read.sanitise_query("SELECT * FROM users WHERE name = ?")

            with sql_adaptor() as connection:
                user_record = read.query(connection, sql_query=query, values=(user_name))


    """
    return _read._query(connection, query=sql_query, values=values)
